local LootBuddy = LibStub("AceAddon-3.0"):NewAddon("LootBuddy", "AceConsole-3.0", "AceEvent-3.0", "AceConsole-3.0");
local L = LibStub('AceLocale-3.0'):GetLocale('LootBuddy', true)

local options = {
	name = "LootBuddy",
	handler = LootBuddy,
	type = "group",
	args = {
		c = {
			name = "Settings",
			type = "group",
			inline = true,
			args = {
				fenabled = {
					name = 'Trashing Enabled',
					desc = "Enable LootBuddy filtering on loot items.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 1,
				},
				debug_mode = {
					name = 'Debug Mode',
					desc = "Loot will not be trashed, you'll receive notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 2,
				},
				silent_mode = {
					name = 'Silent Mode',
					desc = "Trashing chat notices.",
					type = 'toggle',
					set = 'SetOption',
					get = 'GetOption',
					width = "full",
					order = 3,
				},
				filter_bags = {
					name = "Bags",
					type = "group",
					order = 4,
					args = {
						bag0 = {
							name = 'Bag 1',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 1,
						},
						bag1 = {
							name = 'Bag 2',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 2,
						},
						bag2 = {
							name = 'Bag 3',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 3,
						},
						bag3 = {
							name = 'Bag 4',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 4,
						},
						bag4 = {
							name = 'Bag 5',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 5,
						},
						bag_all = {
							name = 'All',
							type = 'toggle',
							set = 'SetOption',
							get = 'GetOption',
							width = "half",
							order = 6,
						},
					},
				},
				cfilters = {
					name = 'Current Filters - Malformed filters are automatically removed',
					type = 'input',
					multiline = 14,
					width = 'full',
					desc = 'add, remove and edit filters, one filter per line.',
					set = 'SetFilters',
					get = 'GetFilters',
					order = 5,
				},
				safelist = {
					name = 'Safe List - Items in this list will never be removed',
					type = 'input',
					multiline = 10,
					width = 'full',
					desc = 'comma delimit item names.',
					set = 'SetSafeList',
					get = 'GetSafeList',
					order = 6,
				},
				help_label = {
					name = L["TagHelp"],
					type = 'description',
					order = 7,
					width = 'full',
				},
			}
		}
	}
}

local defaults = { 
	char = {
		fenabled=true,
		debug_mode=true,
		silent_mode=false,
		bag0=true,
		bag1=true,
		bag2=true,
		bag3=true,
		bag4=true,
		bag_all=true,
		safelist="",
		cfilters="",
		dev_debug=false,
	},
}
local loot, groupLoot, bags, filters, trash, safe = {}, {}, {}, {}, {}, {}
local lootPushed, totalLoot, bagUpdates = 0, 0, 0

function LootBuddy:OnInitialize()
	LootBuddy.db = LibStub("AceDB-3.0"):New("LootBuddyFiltersDB", defaults)
	LootBuddy.version = GetAddOnMetadata("LootBuddy", "Version");
	LibStub("AceConfig-3.0"):RegisterOptionsTable("LootBuddy", options)
	LootBuddy.optionsDialog = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LootBuddy")
	LootBuddy:SetFilters('', LootBuddy.db.char.cfilters)
	LootBuddy:SetSafeList('', LootBuddy.db.char.safelist);
end

function LootBuddy:OnEnable()
	print_str(": Version " .. LootBuddy.version)
	LootBuddy:RegisterEvent("BAG_UPDATE")
	LootBuddy:RegisterEvent("LOOT_OPENED")
	LootBuddy:RegisterEvent("ITEM_PUSH")
	LootBuddy:RegisterEvent("START_LOOT_ROLL")
	LootBuddy:RegisterEvent("CHAT_MSG_LOOT")
	
	LootBuddy:RegisterChatCommand("lootbuddy", "OnSlashCommand")
	LootBuddy:RegisterChatCommand("lb", "OnSlashCommand")
	foodList = foodList or {}
	waterList = waterList or {}
end

function LootBuddy:OnDisable()
	LootBuddy:UnregisterEvent("BAG_UPDATE")
	LootBuddy:UnregisterEvent("LOOT_OPENED")
	LootBuddy:UnregisterEvent("ITEM_PUSH")
	LootBuddy:UnregisterEvent("START_LOOT_ROLL")
	LootBuddy:UnregisterEvent("CHAT_MSG_LOOT")
	
	LootBuddy:UnregisterChatCommand("lootbuddy")
	LootBuddy:UnregisterChatCommand("lb")
end

function LootBuddy:OnSlashCommand(input)
	local command, rest = input:match("^(%S*)%s*(.-)$");
	-- Any leading non-whitespace is captured into command;
	-- the rest (minus leading whitespace) is captured into rest.
	if command == "enable" then
		if LootBuddy.db.char.fenabled == false then
			LootBuddy.db.char.fenabled = true
			LootBuddy:RegisterEvent("BAG_UPDATE")
			LootBuddy:RegisterEvent("LOOT_OPENED")
			LootBuddy:RegisterEvent("ITEM_PUSH")
			LootBuddy:RegisterEvent("START_LOOT_ROLL")
			LootBuddy:RegisterEvent("CHAT_MSG_LOOT")
			
			print_str(': enabled.')
		else
			print_str(': is already enabled.')
		end
	elseif command == "disable" then
		if LootBuddy.db.char.fenabled == true then
			LootBuddy.db.char.fenabled = false
			LootBuddy:UnregisterEvent("BAG_UPDATE")
			LootBuddy:UnregisterEvent("LOOT_OPENED")
			LootBuddy:UnregisterEvent("ITEM_PUSH")
			LootBuddy:UnregisterEvent("START_LOOT_ROLL")
			LootBuddy:UnregisterEvent("CHAT_MSG_LOOT")
			
			print_str(': disabled.')
		else
			print_str(': is already disabled.')
		end
	elseif command == "config" then
	-- Handle removing of the contents of rest... to something.   
		InterfaceOptionsFrame_OpenToCategory(LootBuddy.optionsDialog)
	elseif command == "silent" then
		LootBuddy.db.char.silent_mode = true
	elseif command == "test" then
		ProcessRollResult('TEST', "Voshie receives loot: [Chaos Orb].")
	else
	-- If not handled above, display some sort of help message
		print_str(" Syntax: /lootbuddy config, /lootbuddy enable, /lootbuddy disable");
	end
end

function LootBuddy:LOOT_OPENED(event, auto)
	loot, bags = {}, {}
	totalLoot, lootPushed, bagUpdates = 0, 0, 0
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		-- print_str('GetNumLootItems(): ' .. GetNumLootItems())
		local lootIcon, lootName, lootQuantity, rarity, locked = GetLootSlotInfo(1)
		if lootQuantity ~= nil then
			if lootQuantity == 0 then
				totalLoot = GetNumLootItems() - 1
			else
				totalLoot = GetNumLootItems()
			end
		end
		for i=1,GetNumLootItems() do
			local lootIcon, lootName, lootQuantity = GetLootSlotInfo(i)
			if in_table(groupLoot, lootName) == false then
				if lootQuantity ~= nil then
					if lootQuantity > 0 then
						table.insert(loot, {lootName, false, false})
					end
				end
			else
				totalLoot = totalLoot - 1
			end
		end
		groupLoot = {}
	end
end

function LootBuddy:BAG_UPDATE(event, bag)
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		bagUpdates = bagUpdates + 1
		--print_str(': bag_updates: ' .. (bagUpdates/2) .. ', loot len: ' .. table.getn(loot))
		if table.getn(loot) > 0 then
			-- print_str(': BAG_UPDATE loot exists in bag: ' .. bag)
			if in_table(bags, bag) ~= true then
				table.insert(bags, bag)
			end
			if lootPushed >= totalLoot then
				filter_loot()
				lootPushed, totalLoot, bagUpdates = 0, 0, 0
			end
		end
	end
end

function LootBuddy:ITEM_PUSH(self, event, ...)
	if table.getn(loot) > 0 then
		lootPushed = lootPushed + 1
		-- print_str(': items #' .. lootPushed .. ' pushed, loot looted: ' .. lootPushed .. ' total loot: ' .. totalLoot)
	end
end

function LootBuddy:START_LOOT_ROLL(event, rollID, rollTime)
	local link = GetLootRollItemLink(rollID)
	local lootName = GetItemInfo(link)
	table.insert(groupLoot, lootName)
end

function filter_loot()
	if LootBuddy.db.char.fenabled and table.getn(filters) > 0 then
		local name, link, total_slots
		local llen = table.getn(loot)
		for i=1,table.getn(loot) do
			--print_str(': loot#' .. i .. ': ' .. loot[i][1])
			for bag = 0, 4 do
				if in_table(bags, bag) then
					if loot[i][2] ~= true then
						total_slots = GetContainerNumSlots(bag)
						for slot = 1, total_slots do
							link = GetContainerItemLink(bag, slot)
							if link ~= nil then
								name = GetItemInfo(link)
								if name == loot[i][1] then
									onItem(bag, slot, link)
									loot[i][2] = true
									break
								end
							end
						end
					else
						--print_str(' ' .. loot[i][1] .. ' found already')
					end
				else
					--print_str(' bag #' .. bag .. ' not in bags table')
				end
			end
		end
		loot, bags = {}, {}
	end
end

function onItem(bag, slot, link)
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link)
	local stackValue = itemStackCount * itemSellPrice
	local itemColor = string.match(itemLink, "|?c?f?f?(%x*)|")
	local itemId = GetContainerItemID(bag, slot);
	if LootBuddy.db.char.dev_debug then
		print_str(': item: (' .. itemName .. ') type: (' .. itemType .. ') subtype: (' .. itemSubType .. ') id: (' .. itemId .. ')')
	end
	if table.getn(filters) > 0 then
		if LootBuddy.db.char['bag'..bag] then
			if itemType ~= "Quest" and itemRarity < 4 and in_table(safe, itemName) == false then
				for mi,mf in pairs(filters) do
					local flen = table.getn(mf)
					local m = 0
						for index,filter in pairs(mf) do
							if filter[1] == "name" then
								if filter[3] == itemName then
									m = m + 1
								end
							elseif filter[1] == "type" then
								if LootBuddy.db.char.dev_debug then
									print_str(' type: ' .. itemType .. ' looking for: ' .. filter[3])
								end
								if filter[2] == "=" then
									if filter[3] == "Miscellaneous" then
										if LootBuddy.db.char.dev_debug then
											print_str('itemColor: ' .. itemColor)
										end
										if itemColor == "9d9d9d" then
											m = m + 1
											if LootBuddy.db.char.dev_debug then
												print_str('itemColor match ' .. m)
											end
										end
									elseif filter[3] == "food" or filter[3] == "Food" and itemType == "Consumable" then
										if in_table(foodList, itemId) then
											m = m + 1
										end
									elseif filter[3] == "water" or filter[3] == "Water" and itemType == "Consumable" then
										if in_table(waterList, itemId) then
											m = m + 1
										end
									elseif filter[3] == itemType then
										m = m + 1
									end
								elseif filter[2] == "!=" and filter[3] ~= itemType then
									m = m + 1
								end
							elseif filter[1] == "subtype" then
								if filter[2] == "=" and filter[3] == itemSubType then
									if LootBuddy.db.char.dev_debug then
										print_str(' ' .. link .. ' subtype found: ' .. itemSubType .. ' looking for subtype: ' .. filter[3])
									end
									m = m + 1
									if LootBuddy.db.char.dev_debug then
										print_str('subtype match ' .. m)
									end
								elseif filter[2] == "!=" and filter[3] ~= itemSubType then
									m = m + 1
								end
							elseif filter[1] == "quality" then
								if LootBuddy.db.char.dev_debug then
									print_str('quality found: ' .. itemRarity .. ' looking for quality: ' .. filter[3])
								end
								if filter[2] == "=" and tonumber(filter[3]) == itemRarity then
									m = m + 1
									if LootBuddy.db.char.dev_debug then
										print_str('quality match ' .. m)
									end
								elseif filter[2] == "!=" and tonumber(filter[3]) ~= itemRarity then
									m = m + 1
								end
							elseif filter[1] == "value" then
								if itemSellPrice < tonumber(filter[3]) then
									m = m + 1
								end
							elseif filter[1] == "level" and itemLevel < tonumber(filter[3]) then
								m = m + 1
							end
						end
						if LootBuddy.db.char.dev_debug then
							print_str(': ' .. link .. ' needed (' .. flen .. ') found (' .. m .. ')')
						end
						if m == flen then
							if LootBuddy.db.char.debug_mode then
								print_str(': ' .. link .. ' marked to destroy, no action taken - debug active.')
							else
								DeleteContainerItem(bag, slot, link)
								return true
							end
						end
				end
			end
		-- else
		--	print_str(': filtering off on bag: ' .. bag)
		end
	end
end

function DeleteContainerItem(bag, slot, link, filter)
	PickupContainerItem(bag, slot);
	if CursorHasItem() then
		DeleteCursorItem();
	end
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(link);
	if LootBuddy.db.char.silent_mode ~= true then
		print_str(': ' .. link .. ' destroyed.')
	end
end

function LootBuddy:GetOption(info)
	return LootBuddy.db.char[info[#info]]
end

function LootBuddy:SetOption(info, input)
	LootBuddy.db.char[info[#info]] = input
	if info[#info] == 'bag_all' and input == true then
		options.args.c.args.filter_bags.args.bag_all.name = "None"
		LootBuddy:SetOption({'c', 'filter_bags', 'bag0'}, true)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag1'}, true)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag2'}, true)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag3'}, true)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag4'}, true)
	elseif info[#info] == 'bag_all' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy:SetOption({'c', 'filter_bags', 'bag0'}, false)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag1'}, false)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag2'}, false)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag3'}, false)
		LootBuddy:SetOption({'c', 'filter_bags', 'bag4'}, false)
	end
	if info[#info] == 'bag0' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy.db.char.bag_all = false
	end
	if info[#info] == 'bag1' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy.db.char.bag_all = false
	end
	if info[#info] == 'bag2' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy.db.char.bag_all = false
	end
	if info[#info] == 'bag3' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy.db.char.bag_all = false
	end
	if info[#info] == 'bag4' and input == false then
		options.args.c.args.filter_bags.args.bag_all.name = "All"
		LootBuddy.db.char.bag_all = false
	end
	
	if LootBuddy.db.char.bag0 and LootBuddy.db.char.bag1 and LootBuddy.db.char.bag2 and LootBuddy.db.char.bag3 and LootBuddy.db.char.bag4 then
		options.args.c.args.filter_bags.args.bag_all.name = "None"
		LootBuddy.db.char.bag_all = true
	end
end

function LootBuddy:GetFilters(info)
	return LootBuddy.db.char.cfilters
end

function LootBuddy:SetFilters(info, input)
	filters = {}
	LootBuddy.db.char.cfilters = ""
	if input ~= '' then
		local new_filter = {}
		local fstring = split_str(input, "[^\r\n]+", false);
		--if table.getn(fstring) > 0 then
			for i,f in ipairs(fstring) do
				new_filter = {}
				for w in f:gmatch("[%a]+%s[(=)(!=)(<)]+%s\"[%a%d%s&]+[\"$]") do
					if LootBuddy.db.char.dev_debug then
						print_str(': matched filter found: ' .. w)
					end
					table.insert(new_filter, w)
				end
				if table.getn(new_filter) > 0 then
					table.insert(filters, new_filter)
					LootBuddy.db.char.cfilters = LootBuddy.db.char.cfilters .. table.concat(new_filter, " and ") .. "\r"
				end 
			end
			if LootBuddy.db.char.dev_debug then
				print_str('filters found: ' .. table.getn(filters))
			end
			if table.getn( filters ) > 0 then
				for mi,mf in pairs(filters) do
					if LootBuddy.db.char.dev_debug then
					end
					for index,fval in pairs(mf) do
						if LootBuddy.db.char.dev_debug then
							print_str('breaking down filter')
						end
						local pattern = string.explode(fval:gsub("([%a]+)%s([(=)(!=)(<)]+)%s(\"[%a%d%s&]+\")", "%1_%2_%3"), '_')
						if LootBuddy.db.char.dev_debug then
							print_str('pattern length: ' .. table.getn(pattern))
						end
						if table.getn(pattern) == 3 then
							pattern[3] = trim_str(pattern[3], "\"")
							if pattern[1] == "name" then
								filters[mi] = {}
								filters[mi][1] = pattern
								break
							else
								filters[mi][index] = pattern
							end
						else
							if LootBuddy.db.char.dev_debug then
								print_str('rule isnt correct')
							end
						end
					end
				end
			end
		--else
		--	if LootBuddy.db.char.dev_debug then
		--		print_str('filter string == nil')
		--	end
		--end
	else
		if LootBuddy.db.char.dev_debug then
			print_str(': input == nil')
		end
	end
end

function LootBuddy:GetSafeList(info)
	return LootBuddy.db.char.safelist
end

function LootBuddy:SetSafeList(info, input)
	safe = {}
	LootBuddy.db.char.safelist = ""
	if input ~= '' then
		local safeItems = string.explode(input, ",")
		--if table.getn(fstring) > 0 then
		for i,f in ipairs(safeItems) do
			if f ~= '' then
				table.insert(safe, f)
			end
		end
		LootBuddy.db.char.safelist = table.concat(safe, ", ")
	end
end

function LootBuddy:CHAT_MSG_LOOT(event, msg)
	if GetNumSubgroupMembers() > 0 or GetNumGroupMembers() > 0 then
		 ProcessRollResult(event, msg)
	end
end

function ProcessRollResult(event, msg)
	local filtered = msg:gsub("(%a+)%s(%a+)%s(%a+):%s(.-)%.", "%1_%2_%3_%4")
	local wonMsg = string.explode(filtered, '_')
	if table.getn(wonMsg) == 4 then
		if wonMsg[4] ~= nil then
			local lootName = GetItemInfo(wonMsg[4])
			local itemIndex = item_index(groupLoot, lootName)
			if itemIndex ~= false then
				table.remove(groupLoot, itemIndex)
			end
		end
	end
end

function print_str(s)
	DEFAULT_CHAT_FRAME:AddMessage('LootBuddy' .. s, 0.1, 0.7, 1.0);
end

function split_str(s, p, kv)
	t = {}
	for key,value in s:gmatch(p) do
		if kv == true then
			if value == nil then value = "nil" end
			t[key] = value
		else
			table.insert(t, key)
		end
	end
	if table.getn(t) == 0 then table.insert(t, s) end
	return t;
end

function string.explode(str, sep) 
	local pos, t = 1, {}
	if #sep == 0 or #str == 0 then return end
	for s, e in function() return string.find(str, sep, pos) end do
		table.insert(t, string.trim(string.sub(str, pos, s-1)))
		pos = e+1
	end
	table.insert(t, string.trim(string.sub(str, pos)))
	return t
end

function trim_str(s, tchar)
  return (s:gsub("^" .. tchar .. "*(.-)" .. tchar .. "*$", "%1"))
end

function in_loot_table( lootName )
	for i=1,table.getn(loot) do
		if loot[i][1] == lootName then
			return i
		end
	end
	return -1
end

function in_table(t, sv)
	for _,v in pairs(t) do
		if v == sv then
			return true
		end
	end
	return false
end

function item_index(haystack, needle)
	for i=1,table.getn(haystack) do
		if haystack[i] == needle then
			return i
		end
	end
	return false
end